((Drupal, drupalSettings, $) => {

  'use strict';

  if (!('RefreshLess' in Drupal)) {
    Drupal.RefreshLess = {};
  }

  if (!('classes' in Drupal.RefreshLess)) {
    Drupal.RefreshLess.classes = {};
  }

  /**
   * RefreshLess Turbo behaviours attach and detach class.
   */
  Drupal.RefreshLess.classes.TurboBehaviours = class {

    /**
     * The context element to attach to; usually the <html> element.
     *
     * @type {HTMLElement}
     */
    #context;

    /**
     * Array of <script> element Promises, each settled when they load or fail.
     *
     * @type {Promise[]}
     */
    #scriptPromises = [];

    constructor(context) {

      this.#context = context;

      this.#bindEventHandlers();

    };

    /**
     * Bind all of our event handlers.
     */
    #bindEventHandlers() {

      $(this.#context).on({
        'refreshless:before-scripts-merge': (event) => {
          this.#beforeMergeHandler(event);
        },
      });

    }

    /**
     * 'refreshless:before-scripts-merge' event handler.
     *
     * @param {jQuery.Event} event
     */
    #beforeMergeHandler(event) {

      for (const element of event.detail.new) {

        // Ignore any <script> elements that don't have a 'src' attribute so
        // that we don't wait for something that won't ever load.
        //
        // Note that this also ignores the drupalSettings JSON element so we
        // don't need a separate check for that.
        if (typeof $(element).attr('src') === 'undefined') {
          continue;
        }

        this.#createScriptPromise(element);

      }

    }

    /**
     * Create a Promise that resolves or rejects when a <script> loads/errors.
     *
     * This creates a Promise for the <script> element with the resolve() and
     * reject() callbacks assigned as the load and error event handlers,
     * respectively, and pushes that Promise onto this.#scriptPromises.
     *
     * @param {HTMLScriptElement} element
     *   A <script> element add a load/error Promise for.
     */
    #createScriptPromise(element) {

      this.#scriptPromises.push(new Promise((resolve, reject) => {

        $(element).one({'load': resolve, 'error': reject});

      }));

    }

    /**
     * Attach behaviours.
     *
     * @return {Promise}
     *   A Promise that fulfills when any new scripts have finished loading or
     *   failed to load, and Drupal.attachBehaviors() has been called.
     */
    async attach() {

      // Only when all of the Promises have settled - that is they've either
      // resolved or rejected, i.e. loaded or failed to load - do we attach
      // behaviours. We don't know ahead of time if one or more behaviours
      // being added by one of the newly added scripts depends on something
      // from one of the other newly added scripts, and because the scripts are
      // not being evalutated at first load in a predictable sequence but
      // dynamically later, this reduces the chances of undefined or reference
      // errors.
      await Promise.allSettled(this.#scriptPromises);

      // While not strictly necessary since already resolved Promises that are
      // passed to Promise.allSettled() should just instantly count as settled
      // to that method, cleaning up and letting garbage collection do its thing
      // is preferable here.
      this.#scriptPromises = [];

      Drupal.attachBehaviors(this.#context, drupalSettings);

      const attachEvent = new CustomEvent('refreshless:attach', {
        // Since we're dispatching on an abitrary context element.
        bubbles: true,
      });

      this.#context.dispatchEvent(attachEvent);

      return Promise.resolve();

    };

    /**
     * Detach behaviours.
     *
     * @return {Promise}
     *   A Promise that fulfills when Drupal.detachBehaviors() has been called.
     */
    detach() {

      Drupal.detachBehaviors(this.#context, drupalSettings, 'unload');

      const detachEvent = new CustomEvent('refreshless:detach', {
        // Since we're dispatching on an abitrary context element.
        bubbles: true,
      });

      this.#context.dispatchEvent(detachEvent);

      return Promise.resolve();

    };

  }

})(Drupal, drupalSettings, jQuery);
