((
  html, Drupal,
  TurboBehaviours, TurboDrupalSettings, TurboScriptManager,
  TurboStylesheetManager,
  drupalSettings, $,
) => {

  'use strict';

  if (!('RefreshLess' in Drupal)) {
    Drupal.RefreshLess = {};
  }

  if (!('classes' in Drupal.RefreshLess)) {
    Drupal.RefreshLess.classes = {};
  }

  /**
   * RefreshLess Turbo adapter class.
   */
  Drupal.RefreshLess.classes.Turbo = class {

    /**
     * The context element to attach to; usually the <html> element.
     *
     * @type {HTMLElement}
     */
    #context;

    /**
     * Behaviours class instance; responsible for attaching/detaching.
     */
    #Behaviours;

    /**
     * drupalSettings updater class instance.
     */
    #DrupalSettings;

    /**
     * Script manager class instance.
     */
    #ScriptManager;

    /**
     * Stylesheet manager class instance.
     */
    #StylesheetManager;

    /**
     * Whether the current response was a redirect.
     *
     * Turbo doesn't provide a clean way to detect this in a
     * 'turbo:before-render' or 'turbo:render' event listener so we need to
     * track whether the most recent response was the result of a redirect to
     * avoid triggering a double set of behaviour detach, settings update, and
     * behaviour attach.
     *
     * @type {Boolean}
     *
     * @see https://www.drupal.org/project/refreshless/issues/3397466
     *   Documents the problem and potential solutions.
     *
     * @see this.#skipRender
     */
    #isRedirected = false;

    /**
     * Wether to skip the next render, i.e. detach, settings update, and attach.
     *
     * We need a second flag in addition to this.#isRedirected to ensure that we
     * do still detach behaviours even on a redirect navigation, inelegant
     * though it may be.
     *
     * @type {Boolean}
     *
     * @see this.#isRedirected
     */
    #skipRender = false;

    /**
     * Flag indicating whether refreshless:load event is initial load.
     *
     * I.e. this is only true on the first triggering of that event and then
     * false thereafter.
     *
     * @type {Boolean}
     */
    #initialLoad = true;

    constructor(
      context, Behaviours, DrupalSettings, ScriptManager, StylesheetManager,
    ) {

      this.#context = context;

      this.#Behaviours = new Behaviours(this.#context);

      this.#DrupalSettings = new DrupalSettings(this.#context);

      this.#ScriptManager = new ScriptManager(this.#context);

      this.#StylesheetManager = new StylesheetManager(this.#context);

      this.#bindEventHandlers();

    }

    /**
     * Bind all of our event handlers.
     */
    #bindEventHandlers() {

      $(this.#context).on({
        'turbo:before-render': async (event) => {
          await this.#beforeRenderHandler(event);
        },
        'turbo:render': (event) => {
          this.#renderHandler(event);
        },
        'turbo:before-fetch-request': (event) => {
          this.#beforeFetchHandler(event);
        },
        'turbo:load': (event) => {
          this.#loadHandler(event);
        },
        'turbo:reload': (event) => {
          this.#reloadHandler(event);
        },
        'turbo:before-prefetch': (event) => {
          this.#beforePrefetchHandler(event);
        },
        'turbo:click': (event) => {
          this.#clickHandler(event);
        },
      });

    }

    /**
     * 'turbo:load' event handler.
     *
     * @param {jQuery.Event} event
     *   The event object generated by Turbo.
     */
    #loadHandler(event) {

      const loadEvent = new CustomEvent(
        'refreshless:load', {
          detail: $.extend(true, {}, event.detail, {
            initial: this.#initialLoad,
          }),
        },
      );

      this.#context.dispatchEvent(loadEvent);

      this.#initialLoad = false;

    }

    /**
     * 'turbo:before-prefetch' event handler.
     *
     * @param {jQuery.Event} event
     *   The event object generated by Turbo.
     */
    #beforePrefetchHandler(event) {

      const beforePrefetchEvent = new CustomEvent(
        'refreshless:before-prefetch', {
          bubbles:    true,
          cancelable: true,
        },
      );

      event.target.dispatchEvent(beforePrefetchEvent);

      if (beforePrefetchEvent.defaultPrevented === true) {
        event.preventDefault();
      }

    }

    /**
     * 'turbo:click' event handler.
     *
     * @param {jQuery.Event} event
     *   The event object generated by Turbo.
     */
    #clickHandler(event) {

      const clickEvent = new CustomEvent(
        'refreshless:click', {
          bubbles:    true,
          cancelable: true,
          detail:     event.detail,
        },
      );

      event.target.dispatchEvent(clickEvent);

      if (clickEvent.defaultPrevented === true) {
        event.preventDefault();
      }

    }

    /**
     * 'turbo:before-render' event handler.
     *
     * @param {jQuery.Event} event
     *   The event object generated by Turbo.
     *
     * @see this.#beforeRenderTasks()
     */
    async #beforeRenderHandler(event) {

      // Don't do anything if this is a cached preview being shown.
      if (typeof $(event.target).attr('data-turbo-preview') !== 'undefined') {
        return;
      }

      // If told to skip this render, set the variable back to false and return.
      if (this.#skipRender === true) {

        this.#skipRender = false;

        return;

      }

      try {

        await this.#beforeRenderTasks(event);

      } catch (error) {

        console.error(error);

        // If anything threw an exception somewhere in our event handler, ensure
        // that Turbo resumes rendering as a failsafe.
        return await event.detail.resume();

      }

    }

    /**
     * 'turbo:before-render' event tasks.
     *
     * @param {jQuery.Event} event
     *   The event object generated by Turbo.
     */
    async #beforeRenderTasks(event) {

      // Pause Turbo rendering.
      event.preventDefault();

      /**
       * Zero or more render delaying promises.
       *
       * This functions as a stack that multiple event subscribers push on to,
       * with rendering only resuming once all Promises are resolved or
       * rejected.
       *
       * @type {Promise[]}
       */
      let renderDelayPromises = [];

      const beforeRenderEvent = new CustomEvent(
        'refreshless:before-render', {detail: {
          turboEvent: event,
          delay: (executor) => {

            const promise = new Promise(executor);

            renderDelayPromises.push(promise);

            return promise;

          },
        },
      });

      this.#context.dispatchEvent(beforeRenderEvent);

      await Promise.allSettled(renderDelayPromises);

      await this.#Behaviours.detach();

      // Resume Turbo rendering.
      return await event.detail.resume();

    }

    /**
     * 'turbo:reload' event handler.
     *
     * @param {jQuery.Event} event
     *   The event object generated by Turbo.
     */
    #reloadHandler(event) {

      const reloadEvent = new CustomEvent(
        'refreshless:reload', {detail: event.detail},
      );

      this.#context.dispatchEvent(reloadEvent);

    }

    /**
     * 'turbo:render' event handler.
     *
     * @param {jQuery.Event} event
     *   The event object generated by Turbo.
     *
     * @see this.#renderTasks()
     */
    async #renderHandler(event) {

      // Don't do anything if this is a cached preview being shown.
      if (typeof $(event.target).attr('data-turbo-preview') !== 'undefined') {
        return;
      }

      // If this render happens right after a redirect, skip the next render.
      if (this.#isRedirected === true) {

        this.#isRedirected = false;

        this.#skipRender = true;

        return;

      }

      // Ensure an exception being thrown in the render tasks does not break
      // navigation but is still logged as an error.
      try {

        await this.#renderTasks(event);

      } catch (error) {

        console.error(error);

      }

    }

    /**
     * 'turbo:render' event tasks.
     *
     * @param {jQuery.Event} event
     *   The event object generated by Turbo.
     */
    async #renderTasks(event) {

      const renderEvent = new CustomEvent(
        'refreshless:render', {detail: event.detail},
      );

      this.#context.dispatchEvent(renderEvent);

      // Catch any exception thrown by TurboDrupalSettings and log it so that
      // execution continues and behaviours can be attached. While this may
      // lead to some broken things, it's likely a lot less broken than without
      // behaviours.
      try {

        await this.#DrupalSettings.update();

      } catch (error) {

        console.error(error);

      }

      await this.#Behaviours.attach();

    }

    /**
     * 'turbo:before-fetch-request' event handler.
     *
     * @param {jQuery.Event} event
     *   The event object generated by Turbo.
     */
    #beforeFetchHandler(event) {

      // Add a header indicating that this is a RefreshLess request.
      event.detail.fetchOptions.headers[
        drupalSettings.refreshless.headerName
      ] = 'true';

      // Detect if this fetch request is being made to prefetch, triggering an
      // event if so and returning.
      if (
        'X-Sec-Purpose' in event.detail.fetchOptions.headers &&
        event.detail.fetchOptions.headers['X-Sec-Purpose'] === 'prefetch'
      ) {

        const prefetchEvent = new CustomEvent(
          'refreshless:prefetch', {detail: event.detail},
        );

        this.#context.dispatchEvent(prefetchEvent);

        return;

      }

      /**
       * Whether the current request is a form submit request.
       *
       * Note that we're not checking if this is a POST request because GET
       * requests are also made in some cases, notably Views exposed forms when
       * clicking the Reset button.
       *
       * @type {Boolean}
       */
      let isFormSubmit = false;

      /**
       * 'turbo:submit-start' event handler.
       *
       * This sets isFormSubmit to true if triggered, letting us know that this
       * was a form submit fetch request.
       *
       * @param {jQuery.Event} event
       *   The event object generated by Turbo.
       */
      const submitEventHandler = (event) => {

        isFormSubmit = true;

      };

      $(this.#context).on('turbo:submit-start', submitEventHandler);

      /**
       * The request URL object.
       *
       * @type {URL}
       */
      const requestUrl = event.detail.url;

      // Add a one off event listener for the next response to determine if this
      // was a redirect.
      $(this.#context).one('turbo:before-fetch-response', (event) => {

        // Remove the submit event handler so that we don't keep adding copies
        // of it on each fetch request. This is necessary to do this way rather
        // than by attaching it with .one() because if this fetch request
        // wasn't a form submit, the event handler won't be removed because it
        // wouldn't be triggered at all until a later form submit.
        $(this.#context).off('turbo:submit-start', submitEventHandler);

        /**
         * The response URL object.
         *
         * @type {URL}
         */
        const responseUrl = new URL(
          event.detail.fetchResponse.response.url,
        );

        // Turbo does not perform a double render when submitting a form, so
        // return on that redirect as that would otherwise cause the
        // destination to not have behaviours attached.
        if (isFormSubmit === true) {
          return;
        }

        // If this was not a redirect, trigger the navigation event and return.
        if (
          event.detail.fetchResponse.response.redirected === false
        ) {

          const navigationResponseEvent = new CustomEvent(
            'refreshless:navigation-response', {detail: {
              url: responseUrl,
            }}
          );

          this.#context.dispatchEvent(navigationResponseEvent);

          return;

        }

        const redirectEvent = new CustomEvent(
          'refreshless:redirect', {
            detail: $.extend({}, event.detail, {
              from: requestUrl,
              to:   responseUrl,
            }),
          },
        );

        this.#context.dispatchEvent(redirectEvent);

        this.#isRedirected = true;

      });

    }

  }

  if (!('Turbo' in Drupal.RefreshLess)) {

    Drupal.RefreshLess.Turbo = new Drupal.RefreshLess.classes.Turbo(
      html, TurboBehaviours, TurboDrupalSettings, TurboScriptManager,
      TurboStylesheetManager,
    );

  }

})(
  document.documentElement,
  Drupal,
  Drupal.RefreshLess.classes.TurboBehaviours,
  Drupal.RefreshLess.classes.TurboDrupalSettings,
  Drupal.RefreshLess.classes.TurboScriptManager,
  Drupal.RefreshLess.classes.TurboStylesheetManager,
  drupalSettings,
  jQuery,
);
